mod etc;
mod boot;

use std::fmt::{Debug, Display, Formatter};
use std::fs::File;
use std::io::{Read};
use std::path::Path;
use lazy_static::lazy_static;
use serde::de::DeserializeOwned;
use simple_rw_global::tokio::AsyncGlobalContainer;
use crate::cfg::boot::BootConfigDef;
use crate::cfg::etc::EtcConfigDef;

const INIT_ETC_CFG_FILENAME:&str = "/nagae/elip4ng/config/init.toml";
const INIT_BOOT_CFG_FILENAME:&str = "/boot/config/init.toml";

lazy_static! {
    pub static ref CFG : AsyncGlobalContainer<ConfigManager> = AsyncGlobalContainer::new();
}

pub struct ConfigManager {
    boot_cfg: BootConfigDef,
    etc_cfg: EtcConfigDef,
}

impl ConfigManager {
    pub fn empty() -> ConfigManager{
        let etc_cfg = EtcConfigDef::default();
        let boot_cfg = etc_cfg.get_boot_default_clone();
        ConfigManager{ etc_cfg, boot_cfg }
    }

    pub fn load() -> ConfigManager{
        println!("Loading config from '{}'...", INIT_ETC_CFG_FILENAME);
        let etc_res = ConfigManager::load_file::<EtcConfigDef>(INIT_ETC_CFG_FILENAME);
        let etc_cfg = match etc_res {
            Ok(s) => {s}
            Err(err) => {
                ConfigManager::print_cfg_fail_info(INIT_ETC_CFG_FILENAME, err);
                EtcConfigDef::default()
            }
        };
        println!("Loading config from '{}'...", INIT_BOOT_CFG_FILENAME);
        let boot_res = ConfigManager::load_file::<BootConfigDef>(INIT_BOOT_CFG_FILENAME);
        let boot_cfg = match boot_res {
            Ok(s) => {s}
            Err(err) => {
                ConfigManager::print_cfg_fail_info(INIT_BOOT_CFG_FILENAME, err);
                etc_cfg.get_boot_default_clone()
            }
        };
        ConfigManager{etc_cfg, boot_cfg}
    }

    fn load_file<T>(fname: &str) -> Result<T, CfgError>
        where T: DeserializeOwned,
    {
        let fpath = Path::new(fname);
        let mut file = File::open(fpath)?;
        let mut toml_str = String::new();
        file.read_to_string(&mut toml_str)?;
        let etc = toml::from_str(&toml_str)?;
        Ok(etc)
    }

    fn print_cfg_fail_info(cfgname: &str, err: CfgError) {
        println!("[ERROR] config '{}' load failed: {:?}", cfgname, err);
        println!("Use default setting instead, but some functions may not be available.");
        println!("If you are a user:\n\tPlease contact the device manufacturer for help.");
        println!("If you are the manufacturer:\n\tPlease use maintenance shell looking for detail.");
    }

    pub fn get_etc_cfg(&self) -> &EtcConfigDef {
        &self.etc_cfg
    }

    pub fn get_boot_cfg(&self) -> &BootConfigDef {
        &self.boot_cfg
    }
}

pub async fn init_config_manager() {
    let cm = ConfigManager::load();
    //let cm = ConfigManager::empty();
    CFG.set(cm).await;
}

pub enum CfgError{
    IOError(std::io::Error),
    DeserializeError(toml::de::Error),
}

impl Debug for CfgError {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            CfgError::IOError(e) => {
                write!(f, "IOError: {:?}", e)
            }
            CfgError::DeserializeError(e) => {
                write!(f, "DeserializeError: {:?}", e)
            }
        }
    }
}

impl Display for CfgError {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        (self as &dyn Debug).fmt(f)
    }
}

impl std::error::Error for CfgError {
}

impl From<std::io::Error> for CfgError {
    fn from(value: std::io::Error) -> Self {
        CfgError::IOError(value)
    }
}

impl From<toml::de::Error> for CfgError {
    fn from(value: toml::de::Error) -> Self {
        CfgError::DeserializeError(value)
    }
}
