use std::fmt::{Display, Formatter};
use tokio::io::{AsyncWriteExt, Stderr};
use tokio::sync::mpsc::{channel, Sender, Receiver};

use std::sync::atomic::{AtomicBool, Ordering};
use lazy_static::lazy_static;
use crate::cfg;

lazy_static!{
    static ref DEBUG_MODE: AtomicBool = AtomicBool::new(false);
}

pub trait LoggerWriterSpawner {
    fn get_writer(&self, prefix: &str) -> LogWriter;
}

pub struct Logger {
    tx: Sender<String>,
    rx: Receiver<String>,
    out: Stderr
}

impl Logger {
    pub fn new() -> Logger {
        let (tx, rx) = channel(1);
        let out = tokio::io::stderr();
        Logger{
            tx,
            rx,
            out,
        }
    }

    pub async fn printing(&mut self) {
        loop {
            let s = self.rx.recv().await;
            match s {
                None => {}
                Some(t) => {
                    let _ = self.out.write_all(t.as_bytes()).await;
                }
            }
        }
    }

    pub fn get_spawner(&self) -> LoggerWriterSpawnerImpl {
        LoggerWriterSpawnerImpl{tx: self.tx.clone()}
    }
}

impl LoggerWriterSpawner for Logger {
    fn get_writer(&self, prefix: &str) -> LogWriter {
        LogWriter{tx: self.tx.clone(), prefix: prefix.to_string() }
    }
}

pub struct LogWriter {
    prefix: String,
    tx: Sender<String>,
}

impl LogWriter {
    async fn log(&self, lv: LogLevel, msg: &str) {
        let _ = self.tx.send(format!("[LOG] <{}> ({}) {}\n", lv, self.prefix.as_str(), msg)).await;
    }

    #[inline]
    pub async fn info(&self, msg: &str) {
        self.log(LogLevel::Info, msg).await;
    }

    #[inline]
    pub async fn warn(&self, msg: &str) {
        self.log(LogLevel::Warn, msg).await;
    }

    #[inline]
    pub async fn error(&self, msg: &str) {
        self.log(LogLevel::Error, msg).await;
    }

    #[inline]
    pub async fn fatal(&self, msg: &str) {
        self.log(LogLevel::Fatal, msg).await;
    }

    #[inline]
    pub async fn debug(&self, msg: &str) {
        if DEBUG_MODE.load(Ordering::Relaxed) {
            self.log(LogLevel::Debug, msg).await;
        }
    }

    pub fn is_debug_mode() -> bool {
        DEBUG_MODE.load(Ordering::Relaxed)
    }
}

pub struct LoggerWriterSpawnerImpl{
    tx: Sender<String>
}

impl LoggerWriterSpawner for LoggerWriterSpawnerImpl{
    fn get_writer(&self, prefix: &str) -> LogWriter {
        LogWriter{tx: self.tx.clone(), prefix: prefix.to_string() }
    }
}

enum LogLevel {
    Debug,
    Info,
    Warn,
    Error,
    Fatal,
}

impl Display for LogLevel {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        let s = match self {
            LogLevel::Info => "I",
            LogLevel::Warn => "W",
            LogLevel::Error => "E",
            LogLevel::Fatal => "F",
            LogLevel::Debug => "D",
        };
        write!(f, "{}", s)
    }
}

pub fn set_debug_mode(mode: bool) {
    DEBUG_MODE.store(mode, Ordering::Relaxed);
}

pub async fn set_debug_mode_by_cfg() {
    let cfg_inst = cfg::CFG.get().await;
    let cfg_boot = cfg_inst.get_boot_cfg();
    DEBUG_MODE.store(cfg_boot.yukari_log_debug, Ordering::Relaxed);
}